import json
from contextlib import suppress
from decimal import Decimal

import click
import re
from openpyxl import Workbook, load_workbook
from openpyxl.worksheet.worksheet import Worksheet
from openpyxl.writer.excel import save_virtual_workbook
from typing import List, Optional, Dict, Callable, Union, Tuple, IO, Any

from rossum.schema import transform

SCHEMA_SHEET_NAME = "Schema"
OPTIONS_SHEET_PREFIX = "Options of "

ATTRS_INDEX = 4

DEFAULT_ATTRIBUTE_TYPES = {
    "label": "string",
    "type": "string",
    "rir_field_names": "json",
    "format": "string",
    "default_value": "json",
    "constraints": "json",
    "hidden": "bool",
    "can_export": "bool",
    "score_threshold": "number",
    "width": "number",
    "width_chars": "number",
    "icon": "string",
    "min_occurrences": "number",
    "max_occurrences": "number",
    "can_collapse": "bool",
    "use_rir_content": "bool",
}

HEADER_RE = re.compile(r"(.*) \((.*)\)")

CATEGORIES = ["section", "multivalue", "tuple", "datapoint"]

SchemaContent = List[dict]
SchemaDatapoint = Dict[str, Union[str, int, float, List, Dict]]


class SchemaToXlsx:
    def convert(self, schema_content: SchemaContent) -> bytes:
        workbook = Workbook()
        self._create_main_sheet(workbook, schema_content)
        self._create_option_sheets(workbook, schema_content)
        return save_virtual_workbook(workbook)

    def _create_main_sheet(self, workbook: Workbook, schema_content: SchemaContent) -> None:
        attribute_types = self._create_attribute_types(schema_content)

        schema_sheet = workbook.active
        schema_sheet.title = SCHEMA_SHEET_NAME
        headers = ["Section", "Multivalue", "Tuple", "Data"] + [
            f"{name} ({type_})" for name, type_ in attribute_types.items()
        ]
        schema_sheet.append(headers)
        schema_sheet.freeze_panes = "E2"

        attr_rows = _traverse_schema_in_order(
            schema_content, self._create_datapoint_row, types=attribute_types
        )
        for row in attr_rows:
            schema_sheet.append(row)
        _adjust_column_width(schema_sheet)

    def _create_datapoint_row(
        self, datapoint: Dict, parent_containers: List[dict], types: Dict
    ) -> List[Optional[str]]:
        structural_row = self._create_structural_row(datapoint, parent_containers)
        attr_values = self._datapoint_to_excel_row(datapoint, types)
        return structural_row + attr_values

    @staticmethod
    def _create_option_sheets(workbook: Workbook, schema_content: SchemaContent) -> None:
        def _options_to_excel(datapoint: dict, _) -> None:
            if "options" not in datapoint:
                return
            options_sheet = workbook.create_sheet(OPTIONS_SHEET_PREFIX + datapoint["id"])
            options_sheet.append(["Value", "Label"])
            options_sheet.freeze_panes = "A2"
            for item in datapoint["options"]:
                options_sheet.append([item["value"], item["label"]])
            _adjust_column_width(options_sheet)

        _traverse_schema_in_order(schema_content, _options_to_excel)

    def _create_attribute_types(self, schema_content: SchemaContent) -> Dict[str, str]:
        additional_attributes = _uniq(
            _traverse_schema_in_order(schema_content, self._extract_unknown_attributes)
        )
        if additional_attributes:
            click.echo(f"Warning: unknown attributes: {', '.join(additional_attributes)}", err=True)
        return {**DEFAULT_ATTRIBUTE_TYPES, **additional_attributes}

    @staticmethod
    def _extract_unknown_attributes(datapoint: Dict, _) -> Dict:
        result = {}
        for attr_name in datapoint:
            if attr_name in ("id", "category", "options", "children"):
                continue
            if attr_name not in DEFAULT_ATTRIBUTE_TYPES:
                result[attr_name] = "json"
        return result

    @staticmethod
    def _create_structural_row(
        datapoint: Dict, parent_containers: List[dict]
    ) -> List[Optional[str]]:
        structural_datapoints = parent_containers + [
            {"category": datapoint["category"], "id": datapoint["id"]}
        ]
        row = []
        for category in ("section", "multivalue", "tuple", "datapoint"):
            try:
                id_ = list(filter(lambda i: i["category"] == category, structural_datapoints))[0][
                    "id"
                ]
            except IndexError:
                id_ = None
            row.append(id_)
        return row

    @staticmethod
    def _datapoint_to_excel_row(datapoint: dict, attribute_types: dict) -> List[Optional[str]]:
        def _value_to_excel(key: str):
            try:
                value = datapoint[key]
            except KeyError:
                return None
            if attribute_types[key] == "json":
                return json.dumps(value)
            return value

        return [_value_to_excel(key) for key in attribute_types]


class XlsxToSchema:
    def convert(self, xlsx: IO[bytes]) -> SchemaContent:
        workbook = load_workbook(filename=xlsx)
        schema = self._construct_objects_from_workbook(workbook)
        return self._enhance_datapoints_with_options(workbook, schema)

    def _construct_objects_from_workbook(self, workbook: Workbook) -> SchemaContent:
        schema_sheet = workbook[SCHEMA_SHEET_NAME]
        sheet_rows = ([_safe_strip(cell) for cell in columns] for columns in schema_sheet.values)
        attribute_types = self._extract_attribute_types(next(sheet_rows)[ATTRS_INDEX:])

        schema: SchemaContent = []
        for columns in sheet_rows:
            id_, parent_id, category, attrs = self._extract_attributes(columns)

            schema = transform.traverse_datapoints(
                schema,
                transform.add,
                parent_id=parent_id,
                datapoint_to_add=self._excel_row_to_datapoint(
                    id_, category, attrs, attribute_types
                ),
            )

        return schema

    @staticmethod
    def _extract_attributes(
        columns: List[Optional[str]]
    ) -> Tuple[str, Optional[str], str, List[Optional[str]]]:
        ids, attrs = columns[:ATTRS_INDEX], columns[ATTRS_INDEX:]

        def get_id() -> Tuple[str, Optional[str]]:
            valid_ids: List[str] = [id__ for id__ in ids if id__ is not None]
            id__ = valid_ids[-1]
            try:
                parent_id_: Optional[str] = valid_ids[-2]
            except IndexError:
                parent_id_ = None
            return id__, parent_id_

        def get_category(id__: str) -> str:
            return CATEGORIES[ids.index(id__)]

        id_, parent_id = get_id()
        return id_, parent_id, get_category(id_), attrs

    @staticmethod
    def _enhance_datapoints_with_options(
        workbook: Workbook, schema: SchemaContent
    ) -> SchemaContent:
        for sheet_name in workbook.sheetnames:
            if not sheet_name.startswith(OPTIONS_SHEET_PREFIX):
                continue
            id_ = sheet_name[len(OPTIONS_SHEET_PREFIX) :]
            sheet = workbook[sheet_name]
            sheet_rows = sheet.values
            next(sheet_rows)
            options = [{"value": item[0], "label": item[1]} for item in sheet_rows]
            schema = transform.traverse_datapoints(
                schema, transform.substitute_options, id_=id_, options=options
            )
        return schema

    @staticmethod
    def _extract_attribute_types(attr_headers: List[str]) -> Dict[str, str]:
        result = {}
        for header in attr_headers:
            match = HEADER_RE.match(header)
            if not match:
                raise click.ClickException(f"Invalid header format: {header}")
            result[match.group(1)] = match.group(2)
        return result

    def _excel_row_to_datapoint(
        self, id_: str, category: str, attrs: List[Optional[str]], attribute_types: Dict[str, str]
    ) -> SchemaDatapoint:
        all_attrs = (
            (name, self._deserialize_value(value, type_))
            for value, (name, type_) in zip(attrs, attribute_types.items())
        )
        valid_attrs = {name: value for name, value in all_attrs if value is not None}
        datapoint: SchemaDatapoint = {"id": id_, "category": category, **valid_attrs}
        if datapoint["category"] in ("section", "tuple"):
            datapoint["children"] = []
        elif datapoint["category"] == "multivalue":
            datapoint["children"] = {}
        return datapoint

    @staticmethod
    def _deserialize_value(value: Optional[str], type_: str):
        if value == "" or value is None:
            return None
        if type_ == "string":
            return str(value)
        if type_ == "number":
            try:
                return int(value)
            except ValueError:
                pass
            try:
                return Decimal(value)
            except ValueError:
                raise click.ClickException(f"Cannot convert to number: {value}")
        if type_ == "bool":
            if isinstance(value, bool):
                return value
            try:
                return json.loads(value)
            except (TypeError, ValueError):
                raise click.ClickException(f"Cannot convert to bool: {value}")
        if type_ == "json":
            try:
                return json.loads(value)
            except TypeError:
                raise click.ClickException(f"Cannot convert to json: {value}")
        click.echo(f"Unknown value type: {type_}", err=True)


def _traverse_schema_in_order(
    datapoints: List[dict], callback: Callable, parent_containers: List[Dict] = None, **kwargs
) -> List:
    parent_containers = parent_containers or []
    result = []
    for datapoint in datapoints:
        result.append(callback(datapoint, parent_containers, **kwargs))
        if datapoint["category"] == "datapoint":
            continue
        children = datapoint.get("children", [])
        if not isinstance(children, list):
            children = [children]
        result.extend(
            _traverse_schema_in_order(
                children,
                callback,
                parent_containers + [{"category": datapoint["category"], "id": datapoint["id"]}],
                **kwargs,
            )
        )
    return result


def _uniq(l: List[Dict]) -> Dict[str, str]:
    result: Dict[str, str] = {}
    for item in l:
        result.update(**item)
    return result


def _safe_strip(value: Any) -> Any:
    with suppress(AttributeError):
        return value.strip()
    return value


# Implemented according to https://stackoverflow.com/questions/13197574/openpyxl-adjust-column-width-size
def _adjust_column_width(worksheet: Worksheet) -> None:
    for col in worksheet.columns:
        max_length = 0
        column = col[0].column_letter  # Get the column name
        for cell in col:
            try:  # Necessary to avoid error on empty cells
                if len(str(cell.value)) > max_length:
                    max_length = len(cell.value)
            except KeyError:
                pass
        max_length = min(max_length, 30)
        adjusted_width = (max_length + 2) * 1.2
        worksheet.column_dimensions[column].width = adjusted_width



from dataclasses import dataclass, replace
from typing import Dict, Iterable, List, Optional, Tuple, Union
from . import APIObject


@dataclass(frozen=True)
class Sideload:
    NOT_SET = "NOT_SET"

    plural: str
    singular: str = NOT_SET

    def __post_init__(self) -> None:
        if self.singular is self.NOT_SET:
            # https://docs.python.org/3/library/dataclasses.html#frozen-instances
            object.__setattr__(self, "singular", self.plural.rstrip("s"))

    @staticmethod
    def _parse_query_array(array: str) -> List[str]:
        return [item.strip() for item in array.split(",") if item.strip()]

    def setup_query(self, query: dict) -> None:
        query_sideloads = self._parse_query_array(query.get("sideload", ""))
        if str(self) not in query_sideloads:
            query_sideloads.append(str(self))
        query["sideload"] = ",".join(query_sideloads)

    @staticmethod
    def get_mapping(objects: List[dict]) -> dict:
        return {obj["url"]: obj for obj in objects}

    def __str__(self) -> str:
        return self.plural


@dataclass(frozen=True)
class Content(Sideload):
    schema_ids: Union[Tuple[str], Tuple[()]] = ()

    def setup_query(self, query: dict) -> None:
        if not self.schema_ids:
            # Content sideloading without any schema ID has no effect."
            return
        super().setup_query(query)
        query_schema_ids = self._parse_query_array(query.get("content.schema_id", ""))
        for schema_id in self.schema_ids:
            if schema_id not in query_schema_ids:
                query_schema_ids.append(schema_id)
        query["content.schema_id"] = ",".join(query_schema_ids)

    @staticmethod
    def get_mapping(objects: List[dict]) -> dict:
        mapping: Dict[str, List[dict]] = {}
        for obj in objects:
            mapping.setdefault(obj["url"].rsplit("/", 1)[0], []).append(obj)
        return mapping

    def __call__(self, *schema_ids: Iterable[str]) -> "Content":
        return replace(self, schema_ids=schema_ids)


def to_sideloads(
    sideloads: Optional[Iterable[Union[APIObject, Sideload, str]]]
) -> Iterable[Sideload]:
    def cast(sideload: Union[APIObject, Sideload, str]) -> Sideload:
        if isinstance(sideload, APIObject):
            return Sideload(sideload.plural, sideload.singular)
        elif isinstance(sideload, Sideload):
            return sideload
        elif isinstance(sideload, str):
            return Sideload(sideload)
        else:
            raise TypeError

    return [cast(sideload) for sideload in sideloads or []]


CONTENT = Content("content")



# Module for searching fields in hook payload coming from Rossum
from typing import Optional

import jmespath


def find_single_datapoint(content: list, schema_id: str) -> Optional[dict]:
    datapoint_identifiers = jmespath.search(
        f"[*].children[?schema_id=='{schema_id}'][] | [0]", content
    )
    return datapoint_identifiers


def find_all_line_items_datapoints(content: list) -> list:
    return jmespath.search("[].children[].children[].children[]", content)


def find_line_items_column(content: list, schema_id: str) -> list:
    return jmespath.search(
        f"[].children[].children[].children[?schema_id=='{schema_id}'][]", content
    )


def find_line_items_rows(content: list, schema_id: str) -> Optional[list]:
    return jmespath.search(f"[].children[?schema_id=='{schema_id}'][] | [0] | children[]", content)


def find_multivalue_parent(content: list, schema_id: str) -> Optional[dict]:
    return jmespath.search(f"[*].children[?schema_id=='{schema_id}'][] | [0]", content)


def find_children_of_simple_multivalue(content: list, child_schema_id: str) -> list:
    return jmespath.search(f"[*].children[].children[?schema_id=='{child_schema_id}'][]", content)



import json
from json import JSONDecodeError
from typing import Tuple, Optional, Dict, Any, List

import click
from pathlib import Path
from rossum import option, argument
from rossum.lib import QUEUES
from rossum.lib.api_client import RossumClient
from tabulate import tabulate


def cleanup_config(config: Dict) -> Dict:
    new_config = {}
    for key, value in config.items():
        if key.startswith("config_"):
            new_key = key.replace("config_", "")
            new_config[new_key] = value
    config = {key: value for key, value in new_config.items() if value is not None}
    if "code" in config.keys():
        config["code"] = Path(config["code"]).read_text()
    return config


@click.group("hook")
def cli() -> None:
    pass


@cli.command(name="create", help="Create a hook object.")
@argument.name
@option.hook_type
@option.queue(
    help="Queue IDs, that the hook will be associated with. "
    "Required field - will be assigned to an only queue automatically if not specified."
)
@option.active
@option.events
@option.config_url
@option.config_insecure_ssl
@option.config_secret
@option.config_code
@option.config_runtime
@option.sideload
@option.token_owner
@option.run_after
@option.test
@click.pass_context
def create_command(
    ctx: click.Context,
    name: str,
    hook_type: str,
    queue_ids: Tuple[int, ...],
    active: bool,
    events: Tuple[str, ...],
    sideload: Tuple[str, ...],
    token_owner: Optional[int],
    run_after: List[str] = None,
    test: Optional[str] = None,
    **kwargs,
) -> None:

    with RossumClient(context=ctx.obj) as rossum:
        if not queue_ids:
            queue_urls = [rossum.get_queue()["url"]]
        else:
            queue_urls = []
            for id_ in queue_ids:
                queue_dict = rossum.get_queue(id_)
                if queue_dict:
                    queue_urls.append(queue_dict["url"])

        config = {**kwargs}
        config = cleanup_config(config)

        token_owner_url = f"{rossum.url}/users/{token_owner}" if token_owner else None
        all_run_after_hooks = (
            [f"{rossum.url}/hooks/{hook_id}" for hook_id in run_after] if run_after else []
        )

        try:
            loaded_test = json.loads(test) if test else {}
        except JSONDecodeError:
            click.echo("Could not parse value for --test. Did you pass a valid JSON?")
            return

        response = rossum.create_hook(
            name=name,
            hook_type=hook_type,
            queues=queue_urls,
            active=active,
            events=list(events),
            sideload=list(sideload),
            config=config,
            token_owner=token_owner_url,
            run_after=all_run_after_hooks,
            test=loaded_test,
        )

        additional_fields = [
            value
            for key, value in response["config"].items()
            if key not in ["code", "runtime", "insecure_ssl"]
        ]

        regular_fields = f"{response['id']}, {response['name']}, {response['queues']}, {response['events']}, {response['sideload']}"
        click.echo(
            regular_fields + ", " + f"{', '.join(map(str, additional_fields))}"
            if additional_fields != []
            else regular_fields
        )


@cli.command(name="list", help="List all hooks.")
@click.pass_context
def list_command(ctx: click.Context,):
    with RossumClient(context=ctx.obj) as rossum:
        hooks_list = rossum.get_hooks((QUEUES,))

    headers = ["id", "name", "events", "queues", "active", "sideload"]

    def get_row(hook: dict) -> List[str]:
        fields = [
            hook["id"],
            hook["name"],
            ", ".join(e for e in hook["events"]),
            ", ".join(str(q.get("id", "")) for q in hook["queues"]),
            hook["active"],
            hook["sideload"],
        ]

        additional = ["url", "insecure_ssl", "secret"]

        for field in additional:
            if field in hook["config"]:
                fields.append(hook["config"][field])

        for header in additional:
            if header not in headers:
                headers.append(header)

        hook_list = [item for item in fields]
        return hook_list

    table = [get_row(hook) for hook in hooks_list]

    click.echo(tabulate(table, headers=headers))


@cli.command(name="change", help="Update a hook object.")
@argument.id_
@option.queue(related_object="hook")
@option.name
@option.hook_type
@option.events
@option.active
@option.config_url
@option.config_secret
@option.config_insecure_ssl
@option.config_code
@option.config_runtime
@option.sideload
@option.token_owner
@option.run_after
@option.test
@click.pass_context
def change_command(
    ctx: click.Context,
    id_: int,
    queue_ids: Tuple[int, ...],
    name: Optional[str],
    hook_type: str,
    events: Tuple[str, ...],
    active: Optional[bool],
    sideload: Tuple[str, ...],
    token_owner: Optional[int],
    run_after: List[str] = None,
    test: Optional[str] = None,
    **kwargs,
) -> None:

    config = {**kwargs}
    config = cleanup_config(config)

    if not any([queue_ids, name, active, events, sideload, config]):
        return

    data: Dict[str, Any] = {"config": {}}

    with RossumClient(context=ctx.obj) as rossum:
        if queue_ids:
            data["queues"] = [rossum.get_queue(queue)["url"] for queue in queue_ids]
        if name is not None:
            data["name"] = name
        if hook_type:
            data["type"] = hook_type
        if active is not None:
            data["active"] = active
        if events:
            data["events"] = list(events)
        if config:
            data["config"] = config
        if sideload:
            data["sideload"] = list(sideload)
        if token_owner:
            data["token_owner"] = f"{rossum.url}/users/{token_owner}"
        if run_after:
            data["run_after"] = [f"{rossum.url}/hooks/{hook_id}" for hook_id in run_after]
        if test:
            try:
                loaded_test = json.loads(test)
                data["test"] = loaded_test
            except JSONDecodeError:
                click.echo("Could not parse value for --test. Did you pass a valid JSON?")
                return

        rossum.patch(f"hooks/{id_}", data)


@cli.command(name="delete", help="Delete a hook.")
@argument.id_
@click.confirmation_option(
    prompt="This will delete the hook deployed on the queue. Do you want to continue?"
)
@click.pass_context
def delete_command(ctx: click.Context, id_: str) -> None:
    with RossumClient(context=ctx.obj) as rossum:
        url = rossum.url
        rossum.delete(to_delete={f"{id_}": f"{url}/hooks/{id_}"}, item="hook")



import click

from . import csv_to_options, xls_to_csv, compare


@click.group("tools")
def cli() -> None:
    pass


cli.add_command(compare.cli)
cli.add_command(csv_to_options.cli)
cli.add_command(xls_to_csv.cli)



import json
import functools
from typing import Callable, IO, Optional

import click

from rossum.schema.xlsx import SchemaContent, XlsxToSchema


def schema_content_factory(file_decorator):
    def schema_content(command: Optional[Callable] = None, **file_decorator_kwargs):
        def _load_func(schema_content_file_: Optional[IO[bytes]]) -> Optional[SchemaContent]:
            if schema_content_file_ is None:
                return None

            error = ""
            for name, load_func in (("json", json.load), ("xlsx", XlsxToSchema().convert)):
                try:
                    return load_func(schema_content_file_)  # type: ignore
                except Exception as e:
                    error += f"\n\t{name.upper()}: {e}"
                    schema_content_file_.seek(0)

            raise NotImplementedError(error)

        def decorator(command_: Callable):
            @file_decorator(**file_decorator_kwargs)
            @functools.wraps(command_)
            def wrapped(*args, schema_content_file_: Optional[IO[bytes]], **kwargs):
                try:
                    schema_content_ = _load_func(schema_content_file_)
                except Exception as e:
                    filename = (
                        schema_content_file_.name if schema_content_file_ else schema_content_file_
                    )
                    raise click.ClickException(f"File {filename} could not be loaded.{e}") from e

                return command_(*args, schema_content=schema_content_, **kwargs)

            return wrapped

        if command is None:
            return decorator
        return decorator(command)

    return schema_content



import functools
import json
from typing import IO, Optional, Callable, Iterable

import click

from rossum.common import schema_content_factory
from rossum.lib import split_dict_params

name = click.argument("name", type=str)


def id_(command: Optional[Callable] = None, **kwargs):
    default_kwargs = {"type": int, "metavar": "ID"}
    kwargs = {**default_kwargs, **kwargs}
    decorator = click.argument("id_", **kwargs)
    if command is None:
        return decorator
    return decorator(command)


def schema_file(command: Callable):
    click.argument("schema_file", type=click.File("rb"))(command)

    @functools.wraps(command)
    def wrapped(ctx: click.Context, schema_file: IO[str], *args, **kwargs):
        ctx.obj = {"SCHEMA": json.load(schema_file)}
        return command(ctx, *args, **kwargs)

    return wrapped


def datapoint_parameters(command: Callable):
    click.argument("datapoint_parameters", nargs=-1, type=str)(command)

    @functools.wraps(command)
    def wrapped(*args, datapoint_parameters: Iterable[str], **kwargs):
        try:
            split_params = split_dict_params(datapoint_parameters)
        except ValueError as e:
            raise click.BadArgumentUsage("Expecting <key>=<value> pairs.") from e
        return command(*args, datapoint_parameters=dict(split_params), **kwargs)

    return wrapped


def schema_content_file(command: Optional[Callable] = None, **kwargs):
    default_kwargs = {"type": click.File("rb"), "metavar": "FILE"}
    kwargs = {**default_kwargs, **kwargs}
    decorator = click.argument("schema_content_file_", **kwargs)
    if command is None:
        return decorator
    return decorator(command)


schema_content = schema_content_factory(schema_content_file)



#!/usr/bin/env python3
from typing import Optional, IO

import click as click
import pandas as pd

from rossum import option

HELP = "Create options list from csv. Usable for filling options key of enum datapoint in schema."


@click.command("csv_to_options", help=HELP)
@click.argument("csv", type=click.File("rb"))
@click.option("--delimiter", default=";", type=str)
@click.option("--header", default=None, type=int)
@click.option("--add-value", is_flag=True)
@click.option("--empty-value", default=None, type=str)
@option.output_file
def cli(
    csv: click.File,
    header: Optional[int],
    delimiter: str,
    add_value: bool,
    empty_value: Optional[str],
    output_file: Optional[IO[str]],
) -> None:
    names = ["value", "label"] if header is None else None
    df = pd.read_csv(csv, sep=delimiter, names=names, header=header, dtype=(str, str))
    if add_value:
        df.label = df.apply(": ".join, axis=1)
    if empty_value is not None:
        null_df = pd.DataFrame([[empty_value, "-" * 3]], columns=["value", "label"])
        df = pd.concat([null_df, df])
    click.echo(df.to_json(orient="records").encode("utf-8"), file=output_file, nl=False)



from typing import Optional, Callable

import click

from rossum.common import schema_content_factory


organization = click.option(
    "-o", "--organization-id", type=int, help="Organization ID.", hidden=True
)

name = click.option("-n", "--name", type=str)
email_prefix = click.option(
    "--email-prefix", type=str, help="If not specified, documents cannot be imported via email."
)
bounce_email = click.option(
    "--bounce-email", type=str, help="Unprocessable documents will be bounced to this email."
)
connector_id = click.option(
    "--connector-id", type=str, help="If not specified, queue will not call back a connector."
)

hook_id = click.option(
    "--hook-id",
    type=int,
    multiple=True,
    help="If not specified, hook will not be associated with the queue.",
)

output_file = click.option("-O", "--output-file", type=click.File("wb"))


def schema_content_file(command: Optional[Callable] = None, **kwargs):
    default_kwargs = {"type": click.File("rb"), "help": "Schema file."}
    kwargs = {**default_kwargs, **kwargs}
    decorator = click.option("-s", "--schema-content-file", "schema_content_file_", **kwargs)
    if command is None:
        return decorator
    return decorator(command)


schema_content = schema_content_factory(schema_content_file)


def workspace_id(command: Optional[Callable] = None, **kwargs):
    default_kwargs = {"type": int, "help": "Workspace ID."}
    kwargs = {**default_kwargs, **kwargs}
    decorator = click.option("-w", "--workspace-id", **kwargs)
    if command is None:
        return decorator
    return decorator(command)


def queue(command: Optional[Callable] = None, related_object: Optional[str] = "object", **kwargs):
    default_kwargs = {
        "type": int,
        "help": f"Queue IDs, which the {related_object} will be associated with.",
        "multiple": True,
        "show_default": True,
    }
    kwargs = {**default_kwargs, **kwargs}
    decorator = click.option("-q", "--queue-id", "queue_ids", **kwargs)
    if command is None:
        return decorator
    return decorator(command)


def user(command: Optional[Callable] = None, **kwargs):
    default_kwargs = {
        "type": int,
        "multiple": True,
        "help": "User IDs, which the queues will be associated with.",
    }
    kwargs = {**default_kwargs, **kwargs}
    decorator = click.option("-u", "--user-id", "user_ids", **kwargs)
    if command is None:
        return decorator
    return decorator(command)


service_url = click.option(
    "-u", "--service-url", type=str, required=True, help="Url of the connector endpoint."
)

auth_token = click.option(
    "-t",
    "--auth-token",
    type=str,
    help="Token sent to the connector in the header to ensure authorization. "
    "Generated automatically, if not set manually.",
)

params = click.option("-p", "--params", type=str, help="Query params appended to the service_url.")

asynchronous = click.option(
    "-a", "--asynchronous", type=bool, default=True, help="Affects calling of the connector."
)

active = click.option(
    "--active", type=bool, required=True, default=True, help="Affects whether the hook is notified."
)

events = click.option(
    "-e",
    "--events",
    required=True,
    type=str,
    multiple=True,
    help="List of events, when the hook should be notified.",
)

token_owner = click.option(
    "--token-owner",
    required=False,
    type=int,
    multiple=False,
    help="ID of the user for whom a temporary API token should be generated and passed to the webhook.",
)

run_after = click.option(
    "--run-after",
    required=False,
    type=int,
    multiple=True,
    help="IDs of the hooks that this hook should run after.",
)

test = click.option(
    "--test",
    required=False,
    type=str,
    multiple=False,
    help='Input saved for hook testing purposes. Pass as JSON, e.g. "{\\"config\\": \\"value\\"}"',
)


class OptionRequiredIf(click.Option):
    def full_process_value(self, ctx, value):  # pragma: no cover
        # full_process_value is called in Click < 8.0
        value = super(OptionRequiredIf, self).full_process_value(ctx, value)
        return self._process_value(ctx, value)

    def process_value(self, ctx, value):
        # process_value is called in Click >= 8.0
        value = super(OptionRequiredIf, self).process_value(ctx, value)
        return self._process_value(ctx, value)

    def _process_value(self, ctx, value):
        option = self.human_readable_name

        required_params = {
            "webhook": {
                "mutually_exclusive": "function",
                "config": ["config_url", "config_secret", "config_insecure_ssl"],
            },
            "function": {
                "mutually_exclusive": "webhook",
                "config": ["config_runtime", "config_code"],
            },
        }
        if value is None:
            expected_params = required_params.get(ctx.params["hook_type"], [])
            for opt in expected_params["config"]:
                if option in expected_params["config"]:
                    if opt not in [key for key in ctx.params.keys()]:
                        msg = f"Required if hook type is {ctx.params['hook_type']}"
                        raise click.MissingParameter(ctx=ctx, message=msg)
            for opt in required_params[expected_params["mutually_exclusive"]]["config"]:
                if opt in [key for key in ctx.params.keys()]:
                    if ctx.params.get(opt) is not None:
                        non_valid_param = opt
                        msg = f"Illegal usage: --{non_valid_param} cannot be used for the hook type {ctx.params['hook_type']}"
                        raise click.UsageError(ctx=ctx, message=msg)

        return value


hook_type = click.option(
    "-t",
    "--hook-type",
    required=True,
    type=click.Choice(["function", "webhook"]),
    help="Hook type. Possible values: webhook, function.",
)


config_url = click.option(
    "--config-url",
    type=str,
    help="URL endpoint where the message from the hook should be pushed.",
    cls=OptionRequiredIf,
)

config_secret = click.option(
    "--config-secret",
    type=str,
    help="Secret key for authorization of payloads.",
    cls=OptionRequiredIf,
)

config_insecure_ssl = click.option(
    "--config-insecure-ssl",
    type=bool,
    help="Disable SSL certificate verification. (Use only for testing purposes.)",
    cls=OptionRequiredIf,
)

config_code = click.option(
    "--config-code",
    type=click.Path(readable=True),
    help="Path to the file with the string-serialized source code to be executed.",
    cls=OptionRequiredIf,
)

config_runtime = click.option(
    "--config-runtime", type=str, help="Runtime used to execute code.", cls=OptionRequiredIf
)

sideload = click.option(
    "-s",
    "--sideload",
    default=[],
    type=str,
    multiple=True,
    help="List of related objects that should be included in the hook request.",
)


def group(command: Optional[Callable] = None, **kwargs):
    default_kwargs = {
        "default": "annotator",
        "type": click.Choice(["annotator", "annotator_limited", "admin", "manager", "viewer"]),
        "help": "Permission group.",
        "show_default": True,
    }
    kwargs = {**default_kwargs, **kwargs}
    decorator = click.option("-g", "--group", **kwargs)
    if command is None:
        return decorator
    return decorator(command)


def locale(command: Optional[Callable] = None, **kwargs):
    default_kwargs = {
        "default": "en",
        "type": click.Choice(["en", "cs"]),
        "help": "UI locale",
        "show_default": True,
    }
    kwargs = {**default_kwargs, **kwargs}
    decorator = click.option("-l", "--locale", **kwargs)
    if command is None:
        return decorator
    return decorator(command)


def password(command: Optional[Callable] = None, **kwargs):
    default_kwargs = {"type": str, "required": False, "help": "Generated, if not specified."}
    kwargs = {**default_kwargs, **kwargs}
    if "help" in kwargs and kwargs["help"] is None:
        kwargs.pop("help")
    decorator = click.option("-p", "--password", **kwargs)
    if command is None:
        return decorator
    return decorator(command)



#!/usr/bin/env python3
import operator
from io import StringIO
from typing import Optional

import click as click
import pandas as pd
from typing.io import IO

from rossum import option


@click.command("xls_to_csv", help="Convert an Excel sheet to a CSV. All indices are 0-based.")
@click.argument("xls", type=click.File("rb"))
@click.option("--value", default=0, type=int, help="Index of column with values")
@click.option("--label", default=1, type=int, help="Index of column with labels")
@click.option("--sheet", default=0, type=int, help="Index of sheet")
@click.option(
    "--header",
    default=None,
    type=int,
    help="Index of header row. If not chosen, no header is assumed. "
    "(skiprows are applied before header lookup)",
)
@click.option("--skiprows", default="", type=str, help="Indices of rows to skip")
@option.output_file
def cli(
    xls: click.File,
    label: int,
    value: int,
    sheet: int,
    skiprows: str,
    header: Optional[int],
    output_file: Optional[IO[str]],
) -> None:
    try:
        skiprows_list = [int(r.strip()) for r in skiprows.split(",") if r]
    except ValueError as e:
        raise click.BadArgumentUsage('skiprows: Expecting list of ints delimited by ",".') from e
    cols = {"value": value, "label": label}
    df = pd.read_excel(
        xls,
        sheet_name=sheet,
        usecols=[value, label],
        names=[col for col, _ in sorted(cols.items(), key=operator.itemgetter(1))],
        dtype=(str, str),
        header=header,
        skiprows=skiprows_list,
        engine="openpyxl",
    )
    df = df[list(cols.keys())]
    with StringIO() as buffer:
        df.to_csv(buffer, header=False, sep=";", index=False)
        click.echo(buffer.getvalue().encode("utf-8"), file=output_file, nl=False)



import configparser
import os
from typing import Optional

from pathlib import Path

import click

from rossum import CTX_PROFILE, CTX_DEFAULT_PROFILE

CONFIGURATION_PATH = Path.home() / ".rossum" / "credentials"
DEFAULT_ROSSUM_URL = "https://api.elis.rossum.ai"


ROSSUM_ENV_PROFILE_VAR = "ROSSUM_PROFILE"


HELP = f"""\
Configure API setup.

Credentials are saved into {CONFIGURATION_PATH}.
It is possible to add new or update existing profile by writing down the profile name.
If no profile is chosen, credentials are set to default profile.

Alternatively, configuration can be set using
environment variables:

ROSSUM_URL: URL of the API (e.g. {DEFAULT_ROSSUM_URL})
ROSSUM_USERNAME: username of an ROSSUM account
ROSSUM_PASSWORD: password to the ROSSUM account
"""


@click.command(name="configure", help=HELP)
@click.pass_context
def cli(ctx: click.Context,):
    config = configparser.RawConfigParser()

    if os.path.isfile(CONFIGURATION_PATH):
        with CONFIGURATION_PATH.open("r") as f:
            config.read_file(f)

    config[ctx.obj[CTX_PROFILE]] = {
        "url": click.prompt(f"API URL", default=DEFAULT_ROSSUM_URL, type=str).strip().rstrip("/"),
        "username": click.prompt(f"Username", type=str).strip(),
        "password": click.prompt(f"Password", hide_input=True, type=str).strip(),
    }

    CONFIGURATION_PATH.parent.mkdir(parents=True, exist_ok=True)
    with CONFIGURATION_PATH.open("w") as f:
        config.write(f)


def get_credential(attr: str, profile: Optional[str] = None) -> str:
    res = os.getenv(f"ROSSUM_{attr.upper()}")
    if res is not None:
        return res

    profile = os.getenv(ROSSUM_ENV_PROFILE_VAR) or profile or CTX_DEFAULT_PROFILE

    config = configparser.RawConfigParser()
    config.read(CONFIGURATION_PATH)
    try:
        config_dict = config[profile]
    except KeyError as e:
        raise click.ClickException(
            f"Provide API credential {attr}. "
            f"Either by using `rossum configure`, or environment variable ROSSUM_{attr.upper()}."
        ) from e
    else:
        res = config_dict.get(attr)
    assert res is not None
    return res.strip()



from itertools import chain
from typing import Any, Dict, Optional, Tuple

import click
from tabulate import tabulate

from rossum import argument, option
from rossum.lib import GROUPS, QUEUES, USERS, generate_secret
from rossum.lib.api_client import RossumClient


@click.group("user")
def cli() -> None:
    pass


@cli.command(name="create", short_help="Create user.")
@click.argument("username")
@option.password
@option.queue(related_object="user", required=True)
@option.organization
@option.group
@option.locale
@click.pass_context
def create_command(
    ctx: click.Context,
    username: str,
    password: Optional[str],
    queue_ids: Tuple[int],
    organization_id: Optional[int],
    group: str,
    locale: str,
) -> None:
    """
    Create user with USERNAME and add him to QUEUES specified by ids.
    """
    password = password or generate_secret()
    with RossumClient(context=ctx.obj) as rossum:
        if rossum.get_users(username=username):
            raise click.ClickException(f"User with username {username} already exists.")
        organization = rossum.get_organization(organization_id)

        workspaces = rossum.get_workspaces(organization=organization["id"], sideloads=(QUEUES,))
        queues = chain.from_iterable(w[str(QUEUES)] for w in workspaces)
        queue_urls = [q["url"] for q in queues if q["id"] in queue_ids]

        response = rossum.create_user(
            username, organization["url"], queue_urls, password, group, locale
        )
        click.echo(f"{response['id']}, {password}")


@cli.command(name="list", help="List all users.")
@click.pass_context
def list_command(ctx: click.Context,):
    with RossumClient(context=ctx.obj) as rossum:
        users_list = rossum.get_users((QUEUES, GROUPS), is_active=True)

    table = [
        [
            user["id"],
            user["username"],
            ", ".join(str(g["name"]) for g in user[str(GROUPS)]),
            ", ".join(str(q["id"]) for q in user[str(QUEUES)]),
        ]
        for user in users_list
    ]

    click.echo(tabulate(table, headers=["id", "username", "groups", "queues"]))


@cli.command(name="change", help="Change a user.")
@argument.id_
@option.queue(related_object="user")
@option.group(default=None, show_default=False)
@option.locale(default=None, show_default=False)
@click.pass_context
def change_command(
    ctx: click.Context, id_: int, queue_ids: Tuple[int], group: Optional[str], locale: Optional[str]
) -> None:
    if not any([queue_ids, group, locale]):
        return

    data: Dict[str, Any] = {}

    with RossumClient(context=ctx.obj) as rossum:
        if queue_ids:
            data[str(QUEUES)] = [rossum.get_queue(queue)["url"] for queue in queue_ids]
        if group is not None:
            data[str(GROUPS)] = [g["url"] for g in rossum.get_groups(group_name=group)]
        if locale is not None:
            ui_settings = rossum.get_user(id_)["ui_settings"]
            data["ui_settings"] = {**ui_settings, "locale": locale}

        rossum.patch(f"{USERS}/{id_}", data)


@cli.command(name="delete", help="Delete a user.")
@argument.id_
@click.confirmation_option()
@click.pass_context
def delete_command(ctx: click.Context, id_: int) -> None:
    with RossumClient(context=ctx.obj) as rossum:
        rossum.patch(f"{USERS}/{id_}", {"is_active": False})
